package discoverer_test

import (
	"encoding/json"
	"log"
	"net"
	"strconv"
	"testing"
	"time"

	logging "git.loafle.net/commons/logging-go"
	omd "git.loafle.net/overflow/model/discovery"
	omm "git.loafle.net/overflow/model/meta"
	omu "git.loafle.net/overflow/model/util"
	"git.loafle.net/overflow_scanner/probe/discoverer"
	"github.com/google/gopacket"
	"github.com/google/gopacket/layers"
	"github.com/google/gopacket/pcap"
)

var (
	dz = &omd.DiscoverZone{
		ExcludePatterns: []string{},
	}
	z = &omd.Zone{
		Network:    "192.168.1.0/24",
		MetaIPType: omm.ToMetaIPType(omm.MetaIPTypeEnumV4),
		Address:    "192.168.1.201",
		Iface:      "\\Device\\NPF_{1924FA2B-6927-4BA5-AF43-876C3F8853CE}",
		Mac:        "30:9C:23:15:A3:09",
	}
	dh = &omd.DiscoverHost{
		MetaIPType:     omm.ToMetaIPType(omm.MetaIPTypeEnumV4),
		FirstScanRange: "192.168.1.1",
		LastScanRange:  "192.168.1.254",
	}
	h = &omd.Host{
		Zone:       z,
		MetaIPType: omm.ToMetaIPType(omm.MetaIPTypeEnumV4),
		Address:    "192.168.1.229",
		Mac:        "50:E5:49:46:93:28",
	}
	dp = &omd.DiscoverPort{
		FirstScanRange: 1,
		LastScanRange:  65535,
		ExcludePorts: []int{
			631,
		},
		IncludeTCP: true,
		IncludeUDP: true,
	}
	p = &omd.Port{
		Host:           h,
		MetaPortType:   omm.ToMetaPortType(omm.MetaPortTypeEnumTCP),
		PortNumber:     json.Number(strconv.Itoa(21)),
		DiscoveredDate: omu.NowPtr(),
	}
	ds = &omd.DiscoverService{
		IncludeServices: []string{
			"ACTIVEDIRECTORY",
			"CASSANDRA",
			"DHCP",
			"DNS",
			"FTP",
			"HTTP",
			"IMAP",
			"LDAP",
			"MONGODB",
			"MSSQL",
			"MYSQL",
			"NETBIOS",
			"ORACLE",
			"POP",
			"POSTGRESQL",
			"REDIS",
			"RMI",
			"SMB",
			"SMTP",
			"SNMP",
			"SSH",
			"TELNET",
			"WMI",
			"UNKNOWN",
			"SSH",
			"WMI",
			"SNMP V3",
		},
	}

	dha = &omd.DiscoverHost{
		MetaIPType:     omm.ToMetaIPType(omm.MetaIPTypeEnumV4),
		FirstScanRange: "192.168.1.0",
		LastScanRange:  "192.168.1.255",
		DiscoverPort: &omd.DiscoverPort{
			FirstScanRange: 1,
			LastScanRange:  20000,
			ExcludePorts: []int{
				631,
			},
			IncludeTCP:      true,
			IncludeUDP:      false,
			DiscoverService: ds,
		},
	}
)

func init() {
	// runtime.LockOSThread()
	logging.InitializeLogger("")
}

func TestDiscoverZone(t *testing.T) {
	_discoverer := discoverer.GetDiscoverer()

	handleDiscovery(_discoverer, func(dataChan chan *discoverer.DiscoveryData) {
		_discoverer.DiscoverZone(dataChan, dz)
	})
}

func TestDiscoverHost(t *testing.T) {
	_discoverer := discoverer.GetDiscoverer()

	handleDiscovery(_discoverer, func(dataChan chan *discoverer.DiscoveryData) {
		_discoverer.DiscoverHost(dataChan, z, dh)
	})
}

func TestDiscoverPort(t *testing.T) {
	_discoverer := discoverer.GetDiscoverer()

	handleDiscovery(_discoverer, func(dataChan chan *discoverer.DiscoveryData) {
		_discoverer.DiscoverPort(dataChan, h, dp)
	})
}

func TestDiscoverService(t *testing.T) {
	_discoverer := discoverer.GetDiscoverer()

	handleDiscovery(_discoverer, func(dataChan chan *discoverer.DiscoveryData) {
		_discoverer.DiscoverSerice(dataChan, p, ds)
	})
}

func TestDiscoverHostAll(t *testing.T) {
	_discoverer := discoverer.GetDiscoverer()

	handleDiscovery(_discoverer, func(dataChan chan *discoverer.DiscoveryData) {
		_discoverer.DiscoverHost(dataChan, z, dha)
	})
}

func handleDiscovery(_discoverer discoverer.Discoverer, discoveryFunc func(dataChan chan *discoverer.DiscoveryData)) error {
	var dataChan chan *discoverer.DiscoveryData
	retainChan := make(chan struct{})

	go func() {
		dataChan = _discoverer.Retain()
		close(retainChan)
	}()

	select {
	case <-retainChan:
	}
	defer func() {
		_discoverer.Release(dataChan)
	}()

	go discoveryFunc(dataChan)

	for {
		select {
		case data, ok := <-dataChan:
			if !ok {
				return nil
			}
			switch data.Type {
			case discoverer.DiscoveryDataTypeStart:
				log.Printf("DiscoveryService.DiscoveryStart")
			case discoverer.DiscoveryDataTypeStop:
				log.Printf("DiscoveryService.DiscoveryStop")
				data.Release()
				return nil
			case discoverer.DiscoveryDataTypeError:
				log.Printf("DiscoveryService.DiscoveryDataTypeError %v", data.Error)
			case discoverer.DiscoveryDataTypeZone:
				log.Printf("DiscoveryService.DiscoveryDataTypeZone %v", data.Result)
			case discoverer.DiscoveryDataTypeHost:
				log.Printf("DiscoveryService.DiscoveryDataTypeHost %v", data.Result)
			case discoverer.DiscoveryDataTypePort:
				log.Printf("DiscoveryService.DiscoveryDataTypePort %v", data.Result)
			case discoverer.DiscoveryDataTypeService:
				// log.Printf("DiscoveryService.DiscoveryDataTypeService %v", data.Result)
				log.Println(data.Result)
			}
			data.Release()
		}
	}
}

type scanner struct {
	// iface is the interface to send packets on.
	iface *net.Interface
	// destination, gateway (if applicable), and source IP addresses to use.
	dst, gw, src net.IP

	handle *pcap.Handle

	// opts and buf allow us to easily serialize packets in the send()
	// method.
	opts gopacket.SerializeOptions
	buf  gopacket.SerializeBuffer
}

func TestTCP(t *testing.T) {
	s, err := newScanner()
	if err != nil {
		log.Printf("unable to create scanner for : %v", err)
		return
	}
	if err := s.scan(); err != nil {
		log.Printf("unable to scan %v: %v", err)
	}
	s.close()
}

// newScanner creates a new scanner for a given destination IP address, using
// router to determine how to route packets to that IP.
func newScanner() (*scanner, error) {
	s := &scanner{
		dst: net.ParseIP("192.168.1.229"),
		src: net.ParseIP("192.168.1.201"),
		opts: gopacket.SerializeOptions{
			FixLengths:       true,
			ComputeChecksums: true,
		},
		buf: gopacket.NewSerializeBuffer(),
	}
	// Open the handle for reading/writing.
	// Note we could very easily add some BPF filtering here to greatly
	// decrease the number of packets we have to look at when getting back
	// scan results.
	handle, err := pcap.OpenLive("\\Device\\NPF_{1924FA2B-6927-4BA5-AF43-876C3F8853CE}", 65536, true, pcap.BlockForever)
	if err != nil {
		return nil, err
	}
	s.handle = handle
	return s, nil
}

// close cleans up the handle.
func (s *scanner) close() {
	s.handle.Close()
}

func (s *scanner) scan() error {
	// First off, get the MAC address we should be sending packets to.

	srcMac, err := net.ParseMAC("30:9C:23:15:A3:09")
	if nil != err {
		return err
	}

	dstMac, err := net.ParseMAC("50:E5:49:46:93:28")
	if nil != err {
		return err
	}

	eth := layers.Ethernet{
		SrcMAC:       srcMac,
		DstMAC:       dstMac,
		EthernetType: layers.EthernetTypeIPv4,
	}

	ip4 := layers.IPv4{
		SrcIP:    s.src,
		DstIP:    s.dst,
		Version:  4,
		TTL:      64,
		Protocol: layers.IPProtocolTCP,
	}
	tcp := layers.TCP{
		SrcPort: 60000,
		DstPort: 0, // will be incremented during the scan
		SYN:     true,
		Seq:     0,
	}
	if err := tcp.SetNetworkLayerForChecksum(&ip4); err != nil {
		log.Println(err)
		return err
	}

	// Create the flow we expect returning packets to have, so we can check
	// against it and discard useless packets.
	//ipFlow := gopacket.NewFlow(layers.EndpointIPv4, s.dst, s.src)
	start := time.Now()
	for {
		// Send one packet per loop iteration until we've sent packets
		// to all of ports [1, 65535].
		if tcp.DstPort < 65535 {
			start = time.Now()
			tcp.DstPort++
			if err := s.send(&eth, &ip4, &tcp); err != nil {
				log.Printf("error sending to port %v: %v", tcp.DstPort, err)
			}
		}
		// Time out 5 seconds after the last packet we sent.
		if time.Since(start) > time.Second*60 {
			log.Printf("timed out for %v, assuming we've seen all we can", s.dst)
			return nil
		}

		// Read in the next packet.
		data, _, err := s.handle.ReadPacketData()
		if err == pcap.NextErrorTimeoutExpired {
			continue
		} else if err != nil {
			log.Printf("error reading packet: %v", err)
			continue
		}

		// Parse the packet.  We'd use DecodingLayerParser here if we
		// wanted to be really fast.
		packet := gopacket.NewPacket(data, layers.LayerTypeEthernet, gopacket.NoCopy)
		if tcpLayer := packet.Layer(layers.LayerTypeTCP); tcpLayer == nil {
			continue
		} else if tcp, ok := tcpLayer.(*layers.TCP); !ok {
			continue
		} else if tcp.DstPort != 60000 {
			continue
		} else if tcp.SYN && tcp.ACK {
			log.Println("OPENED: ", tcp.SrcPort)
		}

	}
}

// send sends the given layers as a single packet on the network.
func (s *scanner) send(l ...gopacket.SerializableLayer) error {
	if err := gopacket.SerializeLayers(s.buf, s.opts, l...); err != nil {
		return err
	}
	return s.handle.WritePacketData(s.buf.Bytes())
}
